/* $Workfile: OsaLinuxAsp.c $   $Revision: 4 $    $Modtime: 12/20/07 5:23p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 2007, All Rights Reserved

*******************************************************************************/

/******************************************************************************
  
  Driver exported entry/exit functions.

  Some of the Linux kernel headers don't like to be compiled with g++ (C++)
  so we need a file that is pure C. This file has two purposes: it implements
  the two Linux module specific entry exit routines, init_module and
  cleanup_module, the file operation callbacks, and also some routines that's
  needed for proc operations.

******************************************************************************/
// This is the only place where this file should be included
// It's unlikely it will be needed anywhere else but if so
// __NO_VERSION__ must be defined before inclusion at the
// other places

// Only include with DaVinci
#ifdef DAVINCI


#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/dma-mapping.h>
#include <asm/uaccess.h>
#include <asm/io.h>


// DBG is not defined for C files only CPP
#if OSA_DBG
#define DBG 1
#endif
#include "OsaLinuxRtl.h"  // For OSA_DPRINTF
#if OSA_DBG
#undef DBG
#endif

#include "OsaLinuxAsp.h"
#include "OsaLinuxDmp.h"



//#ifndef DAVINCI_UART0_BASE
//#define OSA_NO_DEV_SIMULATE
//#endif

#undef OSA_NO_DEV_SIMULATE

#ifndef OSA_NO_DEV_SIMULATE
#include <asm/arch/hardware.h>
#include <asm/arch/edma.h>
#include <asm/arch/mcbsp.h>
#else
#define DMA_COMPLETE 1
#endif

#ifndef OSA_NO_DEV_SIMULATE
#define MAX_DMA_SIZE         (0xffff*2)
#else
#define MAX_DMA_SIZE (256)
static struct work_struct dma_emu_work;
static u8 *emu_r_pos;
static u8 *emu_w_pos;
#endif

#define MCBSP_DXR    0x01E02004
#define MCBSP_DRR    0x01E02000
#define MCBSP_SPCR   0x01E02008

static struct proc_dir_entry *devasp_dir;
static struct proc_dir_entry *devasp_regs_dir;

static ssize_t proc_read_reg(struct file *file, char *buf, size_t nbytes, loff_t *ppos);
static ssize_t proc_write_reg(struct file *file, const char *buf, size_t count, loff_t *ppos);
static ssize_t OsaLinuxAsp_simple_read(struct file *file, char *buf, size_t nbytes, loff_t *ppos);
static ssize_t OsaLinuxAsp_simple_write(struct file *file, const char *buf, size_t count, loff_t *ppos);
static int OsaLinuxAsp_release(struct inode *inode, struct file *filp);

static int OsaLinuxAsp_open(struct inode *inode, struct file *filp);
static ssize_t OsaLinuxAsp_write(struct file *file, const char *buffer, size_t count, loff_t *ppos);
static ssize_t OsaLinuxAsp_read(struct file *file, char *buf, size_t nbytes, loff_t *ppos);

DECLARE_WAIT_QUEUE_HEAD(devasp_read_wq);
DECLARE_WAIT_QUEUE_HEAD(devasp_write_wq);

static int OsaLinuxAsp_dma_next_in_buf(void);
static int OsaLinuxAsp_dma_next_out_buf(void);

static unsigned devasp_dma_in_ch_busy;
static int devasp_dma_master_in;
static int devasp_dma_in_ch[2];
static unsigned devasp_dma_next_in_ch;
static unsigned devasp_dma_out_ch_busy;
static int devasp_dma_master_out;
static int devasp_dma_out_ch[2];
static unsigned devasp_dma_next_out_ch;

static unsigned devasp_overruns  = 0;
static unsigned devasp_underruns = 0;
static unsigned devasp_dma_acnt  = 1;
static unsigned devasp_rsyncerr  = 0;
static unsigned devasp_wsyncerr  = 0;

static int devasp_is_open = 0;

#define DEVASP_N_IN_BUFFERS  10
#define DEVASP_N_OUT_BUFFERS 10
#define DEVASP_N_BUFFERS (DEVASP_N_OUT_BUFFERS + 2 + DEVASP_N_IN_BUFFERS + 2)
#define DEVASP_BUF_SIZE OSA_ASP_BUF_SIZE

static struct _devasp_buf {
  void *virt;
  dma_addr_t phys;
  unsigned r_pos;
  unsigned w_pos;
} devasp_buffers[DEVASP_N_BUFFERS];
static struct _devasp_buf *devasp_in_buffers;
static struct _devasp_buf *devasp_out_buffers;
static struct _devasp_buf *devasp_dma_current_in_r_buf;
static struct _devasp_buf *devasp_dma_current_out_w_buf;
static int  devasp_dma_current_in_buf;
static int  devasp_dma_current_out_buf;

static struct inode g_sInode;
static struct file  g_sFileptr;
static struct inode *g_pInodeInit= NULL;
static struct file  *g_pFileInit = NULL;

static struct file_operations proc_reg_operations = {
  read: proc_read_reg,
  write:  proc_write_reg
};

static struct file_operations proc_data_operations = {
  read: OsaLinuxAsp_simple_read,
  write:  OsaLinuxAsp_simple_write
};

static struct file_operations proc_dmadata_operations = {
  read: OsaLinuxAsp_read,
  write:  OsaLinuxAsp_write,
  open: OsaLinuxAsp_open,
  release:  OsaLinuxAsp_release
};

typedef struct asp_reg_entry {
  u32 phyaddr;
  char *name;
  const char *description;
  unsigned phy;
  unsigned int low_ino;
} asp_reg_entry_t;

static asp_reg_entry_t devasp_regs[] = {
 {0x01e02000, "DRR", "Data receive", 1},
 {0x01e02004, "DXR", "Data transmit", 1},
 {0x01e02008, "SPCR", "Serial port control", 1},
 {0x01e0200c, "RCR", "Receive control", 1},
 {0x01e02010, "XCR", "Transmit control", 1},
 {0x01e02014, "SRGR", "Sample rate generator", 1},
 {0x01e02024, "PCR", "Pin control", 1},
 {(u32)&devasp_overruns, "dma_buffer_overruns", "", 0},
 {(u32)&devasp_underruns, "dma_buffer_underruns", "", 0},
 {(u32)&devasp_dma_acnt, "dma_acnt", "", 0},
 {(u32)&devasp_rsyncerr, "rsyncerr", "", 0},
 {(u32)&devasp_wsyncerr, "wsyncerr", "", 0},
};

#ifdef OSA_NO_DEV_SIMULATE
u32 regbuf[100] = {
 0x11111111, /* 0 */
 0x22222222, /* 4 */
 0x33333333, /* 8 */
 0x44444444, /* c */
 0x55555555, /* 10 */
 0x66666666, /* 14 */
 0x77777777, /* 18 */
 0x88888888, /* 1c */
 0x99999999, /* 20 */
 0xaaaaaaaa, /* 24 */
 0xbbbbbbbb, /* 28 */
 0xcccccccc, /* 2c */
};

u8 databuf[MAX_DMA_SIZE] = {
0x00,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0x10,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0x20,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0x30,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0x40,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0x50,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0x60,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0x70,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0x80,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0x90,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0xa0,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0xb0,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0xc0,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0xd0,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0xe0,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
0xf0,1,2,3,4,5,6,7,8,9,0xa,0xb,0xc,0xd, 0xe, 0xf,
};
#endif

#define NUM_OF_REG_ENTRIES (sizeof(devasp_regs)/sizeof(devasp_regs[1]))

#define ASP_MCBSP DAVINCI_MCBSP1


/*
** SPCR = 0x00000000  // In reset
** RCR  = 0x00a113a0
** XCR  = 0x00a013a0
** SRGR = 0x21010000
** PRC  = 0x00000080
*/
static struct davinci_mcbsp_reg_cfg initial_config = {
	.spcr2 = XINTM(0),
	.spcr1 = RINTM(0),
	.rcr2  = RWDLEN2(DAVINCI_MCBSP_WORD_32) | RDATDLY(1),
	.rcr1  = RFRLEN1(3) | RWDLEN1(DAVINCI_MCBSP_WORD_32),
	.xcr2  = XWDLEN2(DAVINCI_MCBSP_WORD_32) | XDATDLY(0),
	.xcr1  = XFRLEN1(3) | XWDLEN1(DAVINCI_MCBSP_WORD_32),
	.srgr1 = FWID(0),
	.srgr2 = CLKSM | FPER(0x101),
	.pcr0  = SCLKME | CLKRP,
};

/******************************************************************************
  proc_read_reg
    return Count
******************************************************************************/
static int 
proc_read_reg(struct file *file, char *buf, size_t nbytes, loff_t *ppos)
{
  int i_ino = (file->f_dentry->d_inode)->i_ino;
  asp_reg_entry_t* current_reg = NULL;
  u32 value;
  char outputbuf[12];
  int count;
  int i;

  if (*ppos > 0) return 0;

  for (i=0; i < NUM_OF_REG_ENTRIES; i++) {
    if (devasp_regs[i].low_ino == i_ino) {
      current_reg = devasp_regs + i;
      break;
    }
  }
  if (current_reg == NULL) {
    return -EINVAL;
  }
  if(current_reg->phy) {
#ifndef OSA_NO_DEV_SIMULATE
    value = *(volatile u32*) io_p2v(current_reg->phyaddr);
#else
    value = regbuf[(current_reg->phyaddr-devasp_regs[0].phyaddr)/4];
#endif
  } else {
    value = *(u32*)current_reg->phyaddr;
  }
  count = sprintf(outputbuf, "0x%08x\n", value);
  if (count > nbytes) return -EINVAL;
  *ppos += count;
  if (copy_to_user(buf, outputbuf, count)) return -EFAULT;
  return count;
}


/******************************************************************************
  proc_write_reg
    return 11
******************************************************************************/
static ssize_t 
proc_write_reg(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
  int i_ino = (file->f_dentry->d_inode)->i_ino;
  asp_reg_entry_t* current_reg = NULL;
  unsigned long value;
  char tmpbuf[12];
  char *endp;
  int i;

  if(count < 11) return -EINVAL;
  if(copy_from_user(tmpbuf, buffer, 11)) return -EFAULT;
  tmpbuf[11] = 0;
  /* 
   * To prevent random values beeing written to reg,
   * strictly check number format.
   */
  if(tmpbuf[10] != '\n' && tmpbuf[10] != ' ') return -EINVAL;
  if(tmpbuf[0] != '0' || (tmpbuf[1] != 'x' && tmpbuf[1] != 'X'))
    return -EINVAL;
  value = simple_strtoul(buffer,&endp,0);
  if(endp-buffer != 10) return -EINVAL;
  
  for(i=0; i < NUM_OF_REG_ENTRIES; i++) {
    if(devasp_regs[i].low_ino==i_ino) {
      current_reg = &devasp_regs[i];
      break;
    }
  }

  if(current_reg == NULL) return -EINVAL;
  if(current_reg->phy) {
#ifndef OSA_NO_DEV_SIMULATE
  *((volatile u32*) io_p2v(current_reg->phyaddr)) = value;
#else
  regbuf[(current_reg->phyaddr-devasp_regs[0].phyaddr)/4] = value;
#endif
  } else {
    *(u32*)current_reg->phyaddr = value;
  }
  
  return 11;
}


/******************************************************************************
  OsaLinuxAsp_simple_write
    return count
******************************************************************************/
static ssize_t 
OsaLinuxAsp_simple_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
  char *tmpbuf;
  int n, dmasize, rc;
  dma_addr_t physaddr;

  tmpbuf = dma_alloc_coherent(NULL, 
                              count > MAX_DMA_SIZE ? MAX_DMA_SIZE : count,
                              &physaddr, GFP_KERNEL);
  if(tmpbuf == NULL) return -ENOMEM;

  n = count;
  while(n > 0) {
    dmasize = n > MAX_DMA_SIZE ? MAX_DMA_SIZE : n;
    if(copy_from_user(tmpbuf, buffer+(count-n), dmasize)) {
      rc = -EFAULT; 
      goto out;
    }
#ifndef OSA_NO_DEV_SIMULATE
    rc = davinci_mcbsp_xmit_buffer(ASP_MCBSP, physaddr, n);
#else
    memcpy(databuf, tmpbuf, n);
    rc = 0;
#endif
    if(rc) goto out;
    n -= dmasize;
  } 
  rc = count;
out:
  dma_free_coherent(NULL, 
                    count > MAX_DMA_SIZE ? MAX_DMA_SIZE : count, 
                    tmpbuf, 
                    physaddr);
  return rc;
}


/******************************************************************************
  OsaLinuxAsp_simple_read
    return count
******************************************************************************/
static ssize_t 
OsaLinuxAsp_simple_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
  char *tmpbuf;
  int n, dmasize, rc;
  dma_addr_t physaddr;

  tmpbuf = dma_alloc_coherent(NULL, count > MAX_DMA_SIZE ? MAX_DMA_SIZE : count,
          &physaddr, GFP_KERNEL);
  if(tmpbuf == NULL) return -ENOMEM;

  n = count;
  while(n > 0) {
    dmasize = n > MAX_DMA_SIZE ? MAX_DMA_SIZE : n;
#ifndef OSA_NO_DEV_SIMULATE
    rc = davinci_mcbsp_recv_buffer(ASP_MCBSP, (dma_addr_t) tmpbuf, n);
#else
    memcpy(tmpbuf, databuf, n);
    rc = 0;
#endif
    if(rc) goto out;
    if(copy_to_user(buffer+(count-n), tmpbuf, dmasize)) {
      rc = -EFAULT; 
      goto out;
    }
    n -= dmasize;
  }
  rc = count;
out:
  dma_free_coherent(NULL, 
                    count > MAX_DMA_SIZE ? MAX_DMA_SIZE : count, 
                    tmpbuf, 
                    physaddr);
  return rc;
}


/******************************************************************************
  OsaLinuxAsp_free_buffers
    return void
******************************************************************************/
static void
OsaLinuxAsp_free_buffers(void)
{
  int i;

  for(i = 0; i < DEVASP_N_BUFFERS; i++) {
    if(devasp_buffers[i].virt != NULL) {
      dma_free_coherent(NULL, 
                        DEVASP_BUF_SIZE, 
                        devasp_buffers[i].virt, 
                        devasp_buffers[i].phys);
    }
  }
}


/******************************************************************************
  OsaLinuxAsp_dma_next_in_buf
    return void
******************************************************************************/
static int
OsaLinuxAsp_dma_next_in_buf(void)
{
  struct _devasp_buf *buf;
  int ch;

  if(devasp_dma_in_ch_busy == 2) {
    return 0; /* this can't happen in irq */
  }
  buf = devasp_in_buffers + devasp_dma_current_in_buf;
  if(buf->r_pos != buf->w_pos) {
    /* do not overwrite data, start transfer to a separate buffer */
    buf = devasp_in_buffers + DEVASP_N_IN_BUFFERS + devasp_dma_next_in_ch;
          devasp_dma_current_in_buf--;
    devasp_overruns++;
  }
  buf->r_pos = buf->w_pos = 0;
  devasp_dma_in_ch_busy++;
  ch = devasp_dma_in_ch[devasp_dma_next_in_ch];
  devasp_dma_next_in_ch = devasp_dma_next_in_ch ^ 1;
#ifndef OSA_NO_DEV_SIMULATE
  davinci_set_dma_src_params(ch, (unsigned long)MCBSP_DRR, 0, 0);
  davinci_set_dma_dest_params(ch, buf->phys, 0, 0);
  davinci_set_dma_src_index(ch, 0, 0);
  davinci_set_dma_dest_index(ch, devasp_dma_acnt, 0);
  davinci_set_dma_transfer_params(ch, 
                                  devasp_dma_acnt, 
                                  DEVASP_BUF_SIZE / devasp_dma_acnt, 
                                  1, 
                                  0, 
                                  ASYNC);
#endif
  devasp_dma_current_in_buf++;
  if(devasp_dma_current_in_buf == DEVASP_N_IN_BUFFERS)
    devasp_dma_current_in_buf = 0;
  return 1;
}


/******************************************************************************
  OsaLinuxAsp_dma_next_out_buf
    return void
******************************************************************************/
static int
OsaLinuxAsp_dma_next_out_buf(void)
{
  struct _devasp_buf *buf;
  int ch;

  if(devasp_dma_out_ch_busy == 2) {
    return 0; /* this can't happen in irq */
  }
  buf = devasp_out_buffers + devasp_dma_current_out_buf;
  if(buf->r_pos == DEVASP_BUF_SIZE) {
    buf->w_pos = buf->r_pos = 0;
    devasp_dma_current_out_buf++;
    if(devasp_dma_current_out_buf == DEVASP_N_OUT_BUFFERS)
      devasp_dma_current_out_buf = 0;
    buf = devasp_out_buffers + devasp_dma_current_out_buf;
  }

  if(buf->r_pos == buf->w_pos) {
    devasp_underruns++;
    buf = devasp_out_buffers + DEVASP_N_OUT_BUFFERS + devasp_dma_next_out_ch;
    buf->r_pos = 0;
    buf->w_pos = DEVASP_BUF_SIZE;
  }

  devasp_dma_out_ch_busy++;
  ch = devasp_dma_out_ch[devasp_dma_next_out_ch];
  devasp_dma_next_out_ch = devasp_dma_next_out_ch ^ 1;
#ifndef OSA_NO_DEV_SIMULATE
  davinci_set_dma_src_params(ch, buf->phys + buf->r_pos, 0, 0);
  davinci_set_dma_dest_params(ch, (unsigned long)MCBSP_DXR, 0, 0);
  davinci_set_dma_src_index(ch, devasp_dma_acnt, 0);
  davinci_set_dma_dest_index(ch, 0, 0);
  davinci_set_dma_transfer_params(ch, 
                                  devasp_dma_acnt, 
                                  (buf->w_pos-buf->r_pos) / devasp_dma_acnt, 
                                  1, 
                                  0, 
                                  ASYNC);
#else
  emu_r_pos = (u8*)buf->virt+buf->r_pos;
  emu_w_pos = (u8*)buf->virt+buf->w_pos;
#endif
  buf->r_pos = buf->w_pos;
  return 1;
}


/******************************************************************************
  OsaLinuxAsp_dma_out_irq
    return void
******************************************************************************/
void 
OsaLinuxAsp_dma_out_irq(int lch, u16 ch_status, void *data)
{
  u32 spcr;

  if(ch_status != DMA_COMPLETE) {
    OSA_EPRINTF(("OsaLinuxAsp: tx dma error, chan=%d, status=0x%x\n", 
                    lch, 
                    ch_status));
    return;
  }
#ifndef OSA_NO_DEV_SIMULATE
  spcr = *(volatile u32*)io_p2v(MCBSP_SPCR);
  if(spcr & (1<<19)) {
    // Bit 19 is XSYNCERR
    *(volatile u32*)io_p2v(MCBSP_SPCR) = spcr & ~(1<<19);
    devasp_wsyncerr++;
  }
#endif
  devasp_dma_out_ch_busy--;
  OsaLinuxAsp_dma_next_out_buf();
  wake_up(&devasp_write_wq);

  OsaLinuxDmpWakeupIrq();

}


/******************************************************************************
  OsaLinuxAsp_dma_in_irq
    return void
******************************************************************************/
void 
OsaLinuxAsp_dma_in_irq(int lch, u16 ch_status, void *data)
{
  unsigned bufno;
  u32 spcr;

  if(ch_status != DMA_COMPLETE) {
    OSA_EPRINTF(("OsaLinuxAsp: dma error, chan=%d, status=0x%x\n", 
                     lch, 
                     ch_status));
    return;
  }
#ifndef OSA_NO_DEV_SIMULATE
  spcr = *(volatile u32*)io_p2v(MCBSP_SPCR);
  if(spcr & (1<<3)) {
    // RSYNCERR bit is bit 3.
    *(volatile u32*)io_p2v(MCBSP_SPCR) = spcr & ~(1<<3);
    devasp_rsyncerr++;
  }
#endif
  bufno = devasp_dma_current_in_buf;
  if(bufno < devasp_dma_in_ch_busy) {
    bufno += DEVASP_N_IN_BUFFERS;
  }
  bufno -= devasp_dma_in_ch_busy;
  devasp_in_buffers[bufno].w_pos = DEVASP_BUF_SIZE;
  devasp_dma_in_ch_busy--;
  OsaLinuxAsp_dma_next_in_buf();
  wake_up(&devasp_read_wq);

  OsaLinuxDmpWakeupIrq();
}


/******************************************************************************
  OsaLinuxAsp_alloc_buffers
    return 1
******************************************************************************/
static int
OsaLinuxAsp_alloc_buffers(void)
{
  int i;
  for(i = 0; i < DEVASP_N_BUFFERS; i++) devasp_buffers[i].virt = NULL;
  for(i = 0; i < DEVASP_N_BUFFERS; i++) {
    devasp_buffers[i].virt = dma_alloc_coherent(NULL, 
                                                DEVASP_BUF_SIZE, 
                                                &devasp_buffers[i].phys, 
                                                GFP_KERNEL);
    if(devasp_buffers[i].virt == NULL) {
      return 0;
    }

    memset(devasp_buffers[i].virt, 0, DEVASP_BUF_SIZE);
    devasp_buffers[i].w_pos = 0;
    devasp_buffers[i].r_pos = 0;
  }
  devasp_in_buffers = devasp_buffers;
  devasp_out_buffers = devasp_buffers + DEVASP_N_IN_BUFFERS + 2;
  devasp_dma_current_in_r_buf = devasp_in_buffers;
  devasp_dma_current_out_w_buf = devasp_out_buffers;
  devasp_dma_current_in_buf = 0;
  devasp_dma_current_out_buf = 0;
  return 1;
}


/******************************************************************************
  OsaLinuxAsp_alloc_rx_dma
    return 1-.OK, 0->error
******************************************************************************/
static int  
OsaLinuxAsp_alloc_rx_dma(void)
{
#ifndef OSA_NO_DEV_SIMULATE   
  int tcc;
  int i;
  int rc;

  rc = davinci_request_dma(DAVINCI_DMA_MCBSP1_RX, 
                           NULL, 
                           OsaLinuxAsp_dma_in_irq, 
                           NULL,
                           &devasp_dma_master_in, 
                           &tcc, 
                           EVENTQ_0);
    
  if(rc != 0) {
     OSA_EPRINTF(("OsaLinuxAsp_alloc_rx_dma: allocate channel: code %d\n", rc));
     return 0;
  }

  for(i = 0; i < 2; i++) {
    rc = davinci_request_dma(DAVINCI_EDMA_PARAM_ANY, 
                             NULL, 
                             OsaLinuxAsp_dma_in_irq, 
                             NULL,
                             devasp_dma_in_ch + i, 
                             &tcc, 
                             EVENTQ_0);
    if(rc != 0) {
     davinci_free_dma(devasp_dma_master_in);
     if(i == 0) davinci_free_dma(devasp_dma_in_ch[0]);
       OSA_EPRINTF(("OsaLinuxAsp_alloc_rx_dma: allocate channel: code %d\n", rc));
     return 0;
    }
  }
  davinci_dma_link_lch(devasp_dma_in_ch[0], devasp_dma_in_ch[1]);
  davinci_dma_link_lch(devasp_dma_in_ch[1], devasp_dma_in_ch[0]);
#endif
  devasp_dma_next_in_ch = 0;
  devasp_dma_in_ch_busy = 0;
  return 1;
}


/******************************************************************************
  OsaLinuxAsp_alloc_tx_dma
    return 1->OK, 0->error
******************************************************************************/
static int  
OsaLinuxAsp_alloc_tx_dma(void)
{
#ifndef OSA_NO_DEV_SIMULATE   
  int tcc;
  int i;
  int rc;

  rc = davinci_request_dma(DAVINCI_DMA_MCBSP1_TX, 
                           NULL, 
                           OsaLinuxAsp_dma_out_irq, 
                           NULL,
                           &devasp_dma_master_out, 
                           &tcc, 
                           EVENTQ_0);
    
  if(rc != 0) {
     OSA_EPRINTF(("OsaLinuxAsp_alloc_tx_dma: allocate channel: code %d\n", rc));
     return 0;
  }

  for(i = 0; i < 2; i++) {
    rc = davinci_request_dma(DAVINCI_EDMA_PARAM_ANY, 
                             NULL, 
                             OsaLinuxAsp_dma_out_irq, 
                             NULL,
                             devasp_dma_out_ch + i, 
                             &tcc, 
                             EVENTQ_0);
    if(rc != 0) {
     davinci_free_dma(devasp_dma_master_out);
     if(i == 0) {
       davinci_free_dma(devasp_dma_out_ch[0]);
     }

     OSA_DPRINTF(("OsaLinuxAsp_alloc_tx_dma: allocate channel: code %d\n", rc));
     return 0;
    }
  }
  davinci_dma_link_lch(devasp_dma_out_ch[0], devasp_dma_out_ch[1]);
  davinci_dma_link_lch(devasp_dma_out_ch[1], devasp_dma_out_ch[0]);
#endif
  devasp_dma_next_out_ch = 0;
  devasp_dma_out_ch_busy = 0;
  return 1;
}


/******************************************************************************
  OsaLinuxAsp_free_rx_dma
    return 1
******************************************************************************/
static void
OsaLinuxAsp_free_rx_dma(void)
{
#ifndef OSA_NO_DEV_SIMULATE
     davinci_free_dma(devasp_dma_master_in);
     davinci_free_dma(devasp_dma_in_ch[0]);
     davinci_free_dma(devasp_dma_in_ch[1]);
#endif
}


/******************************************************************************
  OsaLinuxAsp_free_tx_dma
    return 1
******************************************************************************/
static void
OsaLinuxAsp_free_tx_dma(void)
{
#ifndef OSA_NO_DEV_SIMULATE
     davinci_free_dma(devasp_dma_master_out);
     davinci_free_dma(devasp_dma_out_ch[0]);
     davinci_free_dma(devasp_dma_out_ch[1]);
#endif
}


#ifdef OSA_NO_DEV_SIMULATE
/******************************************************************************
  OsaLinuxAsp_dma_emu_irq
    return 1
******************************************************************************/
void
OsaLinuxAsp_dma_emu_irq(void *a)
{
  unsigned bufno;
  u8 *rp, *wp;
  int i;

  bufno = devasp_dma_current_in_buf;
  if(bufno < devasp_dma_in_ch_busy) bufno += DEVASP_N_IN_BUFFERS;
  bufno -= devasp_dma_in_ch_busy;
  rp = emu_r_pos;
  wp = emu_w_pos;
  emu_r_pos = emu_w_pos = 0;
  for(i = 0; i < DEVASP_BUF_SIZE && rp < wp; i++) {
      ((u8*)devasp_in_buffers[bufno].virt)[i] = *rp++;
  }
  for(; i < DEVASP_BUF_SIZE; i++) {
    ((u8*)devasp_in_buffers[bufno].virt)[i] = 0xff;
  }
  OsaLinuxAsp_dma_out_irq(0, DMA_COMPLETE, NULL);
  OsaLinuxAsp_dma_in_irq(0, DMA_COMPLETE, NULL);
  schedule_delayed_work(&dma_emu_work, usecs_to_jiffies(100000));
}
#endif


/******************************************************************************
  OsaLinuxAsp_dma_start_rx
    return void
******************************************************************************/
static void
OsaLinuxAsp_dma_start_rx(void)
{
#ifndef OSA_NO_DEV_SIMULATE
  edmacc_paramentry_regs r;
#endif

#ifndef OSA_NO_DEV_SIMULATE
  davinci_get_dma_params(devasp_dma_in_ch[devasp_dma_next_in_ch], &r);
  davinci_set_dma_params(devasp_dma_master_in, &r);

  davinci_start_dma(devasp_dma_master_in);
#endif
}


/******************************************************************************
  OsaLinuxAsp_dma_start_tx
    return void
******************************************************************************/
static void
OsaLinuxAsp_dma_start_tx(void)
{
#ifndef OSA_NO_DEV_SIMULATE
  edmacc_paramentry_regs r;
#endif

#ifndef OSA_NO_DEV_SIMULATE
  davinci_get_dma_params(devasp_dma_out_ch[devasp_dma_next_out_ch], &r);
  davinci_set_dma_params(devasp_dma_master_out, &r);

  davinci_start_dma(devasp_dma_master_out);
#endif
}

/******************************************************************************
  OsaLinuxAspOpen
    return 0->OK, Or Error
******************************************************************************/
int 
OsaLinuxAspOpen(void) {
  int rc = 0;

  // Already opened (?)
  if((g_pInodeInit != NULL) || (g_pFileInit != NULL)){
    OSA_DPRINTF(("OsaLinuxAspOpen: device is already open\n"));
    return -EEXIST;
  }

  g_pInodeInit= &g_sInode;
  g_pFileInit = &g_sFileptr;

  rc = OsaLinuxAsp_open(g_pInodeInit, g_pFileInit);

  return rc;
}


/******************************************************************************
  OsaLinuxAsp_open
    return 0->OK, Or Error
******************************************************************************/
static int 
OsaLinuxAsp_open(struct inode *inode, struct file *filp)
{
  /* FIXME: spinlock? */

  int of = devasp_is_open;
  int nf;
  int tx_dma_allocated = 0;
  int rx_dma_allocated = 0;

  if((g_pInodeInit != NULL) || (g_pFileInit != NULL)){
    OSA_EPRINTF(("OsaLinuxAsp_open: not called from OsaLinuxAspOpen\n"));
    g_pInodeInit= inode;
    g_pFileInit = filp;
  }


  nf = filp->f_mode & (FMODE_READ|FMODE_WRITE);
  if((of & nf) != 0) {
    OSA_EPRINTF(("OsaLinuxAsp_open: device is busy\n"));
    return -EBUSY;
  }
  devasp_is_open = of|nf;
  if(of != 0) {
    OSA_EPRINTF(("OsaLinuxAsp_open: device is open\n"));
    return of;
  }

  if(!OsaLinuxAsp_alloc_buffers()) goto fail;
  if(!OsaLinuxAsp_alloc_rx_dma()) goto fail;
  rx_dma_allocated = 1;
  if(!OsaLinuxAsp_alloc_tx_dma()) goto fail;
  tx_dma_allocated = 1;
  OsaLinuxAsp_dma_next_in_buf();
  OsaLinuxAsp_dma_next_in_buf();
  OsaLinuxAsp_dma_next_out_buf();
  OsaLinuxAsp_dma_next_out_buf();
  OsaLinuxAsp_dma_start_tx();
  OsaLinuxAsp_dma_start_rx();
#ifndef OSA_NO_DEV_SIMULATE
  //davinci_mcbsp_start(ASP_MCBSP);
#else
  INIT_WORK(&dma_emu_work, OsaLinuxAsp_dma_emu_irq, NULL);
  schedule_delayed_work(&dma_emu_work, usecs_to_jiffies(100000));
#endif
  return 0;

fail:
  if(tx_dma_allocated) {
    OsaLinuxAsp_free_tx_dma();
  }
  if(rx_dma_allocated) {
    OsaLinuxAsp_free_rx_dma();
  }
  OsaLinuxAsp_free_buffers();
  // If it is not in the open state then the following should be 
  // equal to 0. ???
  devasp_is_open = of;

  return -ENOMEM;
}


/******************************************************************************
  OsaLinuxAsp_release
    return 0->OK, Or Error
******************************************************************************/
static int 
OsaLinuxAsp_release(struct inode *inode, struct file *filp)
{
  int nf = devasp_is_open;
  nf &= ~(filp->f_mode & (FMODE_READ|FMODE_WRITE));

  if(nf != 0) {
    devasp_is_open = nf;
    OSA_EPRINTF(("OsaLinuxAsp_open: device is open\n"));
    return nf;
  }
#ifndef OSA_NO_DEV_SIMULATE
  //davinci_mcbsp_stop(ASP_MCBSP);
  davinci_stop_dma(devasp_dma_master_in);
  davinci_stop_dma(devasp_dma_master_out);
  /* FIXME: wait for dma to complete??? */
#else
  cancel_delayed_work(&dma_emu_work);
#endif
  OsaLinuxAsp_free_tx_dma();
  OsaLinuxAsp_free_rx_dma();
  OsaLinuxAsp_free_buffers();

  devasp_is_open = nf;

  return 0;
}


/******************************************************************************
  OsaLinuxAspRead
    return count, Or Error if negative
******************************************************************************/
ssize_t 
OsaLinuxAspRead(char *buffer, size_t count)
{
  int rc = 0;
  loff_t ppos = 0;

  if((g_pInodeInit == NULL) || (g_pFileInit == NULL)){
    OSA_EPRINTF(("OsaLinuxAsp_open: device was not open\n"));
    rc = -ENXIO;
  }
  else {
    rc = OsaLinuxAsp_read(g_pFileInit, buffer, count, (loff_t *) &ppos);
  }

  return rc;
}


/******************************************************************************
  OsaLinuxAsp_read
    return 0->OK, Or Error
******************************************************************************/
static ssize_t 
OsaLinuxAsp_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
  int rc;
  size_t n, buf_avail, to_read;
  struct _devasp_buf *buf;
  DECLARE_WAITQUEUE(wait, current);

  n = count;
  buf = devasp_dma_current_in_r_buf;

  while(n > 0) {

    if(buf->r_pos == DEVASP_BUF_SIZE) {
      buf++;
      if(buf-devasp_in_buffers == DEVASP_N_IN_BUFFERS)
        buf = devasp_in_buffers;
    }
  
    set_current_state(TASK_INTERRUPTIBLE);
    add_wait_queue(&devasp_read_wq, &wait);
    while(buf->w_pos == 0 && !signal_pending(current)) {
      schedule();
      set_current_state(TASK_INTERRUPTIBLE);
    }
    remove_wait_queue(&devasp_read_wq, &wait);
    set_current_state(TASK_RUNNING);
    if(signal_pending(current)) {
        rc = -EINTR;
        goto out;
    }

    buf_avail = DEVASP_BUF_SIZE - buf->r_pos;
    to_read = buf_avail > n ? n : buf_avail;
    if(copy_to_user(buffer+(count-n), (u8*)buf->virt + buf->r_pos, to_read)) {
      rc = -EFAULT;
      goto out;
    }

    n -= to_read;
    buf->r_pos += to_read;
  }
  rc = count - n;
out:
  devasp_dma_current_in_r_buf = buf;


  return rc;
}


/******************************************************************************
  OsaLinuxAspWrite
    return count, Or Error if negative
******************************************************************************/
ssize_t 
OsaLinuxAspWrite(const char *buffer, size_t count)
{
  int rc = 0;
  loff_t ppos = 0;

  if((g_pInodeInit == NULL) || (g_pFileInit == NULL)){
    OSA_EPRINTF(("OsaLinuxAsp_open: device was not open\n"));
    rc = -ENXIO;
  }
  else {
    rc = OsaLinuxAsp_write(g_pFileInit, buffer, count, &ppos);
  }

  return rc;
}


/******************************************************************************
  OsaLinuxAsp_write
    return count, Or Error if negative
******************************************************************************/
static ssize_t 
OsaLinuxAsp_write(struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
  size_t n, buf_avail, to_write;
  struct _devasp_buf *buf;
  int rc = 0;
  DECLARE_WAITQUEUE(wait, current);

  n = count;
  buf = devasp_dma_current_out_w_buf;

  while(n > 0) {

    if(buf->w_pos == DEVASP_BUF_SIZE) {
      buf++;
      if(buf-devasp_out_buffers == DEVASP_N_OUT_BUFFERS)
        buf = devasp_out_buffers;
    }
  
    set_current_state(TASK_INTERRUPTIBLE);
    add_wait_queue(&devasp_write_wq, &wait);
    while(buf->w_pos == DEVASP_BUF_SIZE && !signal_pending(current)) {
      schedule();
      set_current_state(TASK_INTERRUPTIBLE);
    }
    remove_wait_queue(&devasp_write_wq, &wait);
    set_current_state(TASK_RUNNING);
    if(signal_pending(current)) {
        rc = -EINTR;
        goto out;
    }

    buf_avail = DEVASP_BUF_SIZE - buf->w_pos;
    to_write = buf_avail > n ? n : buf_avail;
    if(copy_from_user((u8*)buf->virt + buf->w_pos, buffer+(count-n), to_write)) {
      rc = -EFAULT;
      goto out;
    }

    n -= to_write;
    buf->w_pos += to_write;
  }
  rc = count - n;
out:
  devasp_dma_current_out_w_buf = buf;
  return rc;
}


/******************************************************************************
  OsaLinuxAsp_remove_proc_entries
    return void
******************************************************************************/
static void
OsaLinuxAsp_remove_proc_entries(void)
{
  int i;

  if(devasp_regs_dir) {
    for(i = 0; i < NUM_OF_REG_ENTRIES; i++) 
      remove_proc_entry(devasp_regs[i].name, devasp_regs_dir);
    remove_proc_entry("registers", devasp_dir);
  }
  remove_proc_entry("data", devasp_dir);
  remove_proc_entry("dmadata", devasp_dir);
  if(devasp_dir) {
    remove_proc_entry("devasp", &proc_root);
  }
}


/******************************************************************************
  init_devasp
    return 0->OK, Or Error
******************************************************************************/
static int __init 
init_devasp(void)
{
  struct proc_dir_entry *entry;
  int i;

  devasp_regs_dir = NULL;

  OSA_DPRINTF(("devasp: Init Started /proc/devasp\n"));

  devasp_dir = proc_mkdir("devasp", &proc_root);
  if (devasp_dir == NULL) {
    OSA_EPRINTF(("devasp: can't create /proc/devasp\n"));
    return -EINVAL;
  }

  devasp_regs_dir = proc_mkdir("registers", devasp_dir);
  if (devasp_regs_dir == NULL) {
    OSA_EPRINTF(("devasp: can't create /proc/devasp/registers\n"));
    goto fail;
  }

  for(i = 0; i < NUM_OF_REG_ENTRIES; i++) {
    entry = create_proc_entry(devasp_regs[i].name,
        S_IWUSR |S_IRUSR | S_IRGRP | S_IROTH, devasp_regs_dir);
    if(entry == NULL) {
      OSA_EPRINTF(("devasp: %s entry == NULL\n", devasp_regs[i].name));
      goto fail;
    }
    devasp_regs[i].low_ino = entry->low_ino;
    entry->proc_fops = &proc_reg_operations;
  }
    
  entry = create_proc_entry("data",
        S_IWUSR |S_IRUSR | S_IRGRP | S_IROTH, devasp_dir);
  if(entry == NULL) {
    OSA_EPRINTF(("devasp: data entry == NULL\n"));
    goto fail;
  }
  entry->proc_fops = &proc_data_operations;
  
  entry = create_proc_entry("dmadata",
        S_IWUSR |S_IRUSR | S_IRGRP | S_IROTH, devasp_dir);
  if(entry == NULL) {
    OSA_EPRINTF(("devasp: dmadata entry == NULL\n"));
    goto fail;
  }
  entry->proc_fops = &proc_dmadata_operations;


#ifndef OSA_NO_DEV_SIMULATE
  //davinci_mcbsp_stop(ASP_MCBSP);

  davinci_mcbsp_config(ASP_MCBSP, &initial_config);

  i = davinci_mcbsp_request(ASP_MCBSP);
  if(i != 0) {
    if(i == -1) {
      OSA_DPRINTF(("devasp: MCBSP1 in use trying to free\n"));
      davinci_mcbsp_free(ASP_MCBSP);
      i = davinci_mcbsp_request(ASP_MCBSP);
      if(i != 0) {
        OSA_EPRINTF(("devasp: can't claim MCBSP1, code %d\n", i));
        goto fail;
      }
    }
    else {
      OSA_EPRINTF(("devasp: can't claim MCBSP1, code %d\n", i));
      goto fail;
    }
  }
#endif

  return 0;

fail:
  OSA_EPRINTF(("devasp: Failed Exit\n"));
  OsaLinuxAsp_remove_proc_entries();
  return -EINVAL;
}


/******************************************************************************
  OsaLinuxAspInitDevAsp
    return 0->OK, Or Error
******************************************************************************/
int  
OsaLinuxAspInitDevAsp(void)
{
  OSA_DPRINTF(("OsaLinuxAspInitDevAsp\n"));

  return init_devasp();
}


/******************************************************************************
  fini_devasp
    return void
******************************************************************************/
static void __exit 
fini_devasp(void)
{
#ifndef OSA_NO_DEV_SIMULATE
  //davinci_mcbsp_stop(ASP_MCBSP);

  davinci_mcbsp_free(ASP_MCBSP);
#endif
  OsaLinuxAsp_remove_proc_entries();
}


/******************************************************************************
  OsaLinuxAspClose
    return void
******************************************************************************/
void  
OsaLinuxAspClose(void)
{

  if((g_pInodeInit == NULL) || (g_pFileInit == NULL)){
    OSA_DPRINTF(("OsaLinuxAspClose: device was not open\n"));
  }
  else {
    OsaLinuxAsp_release(g_pInodeInit, g_pFileInit);
  }

  fini_devasp();

  g_pInodeInit= NULL;
  g_pFileInit = NULL;

  devasp_is_open = 0;
}


/******************************************************************************
  OsaLinuxAspGetStatus 
    return error counters
******************************************************************************/
void  
OsaLinuxAspGetStatus(OsaAspStatusS *status, OsaBool32 resetStatus)
{

  status->devasp_overruns  = devasp_overruns;
  status->devasp_underruns = devasp_underruns;
  status->devasp_dma_acnt  = devasp_dma_acnt;
  status->devasp_rsyncerr  = devasp_rsyncerr;
  status->devasp_wsyncerr  = devasp_wsyncerr;

  // Reset the count to zero
  if(resetStatus) {
    devasp_overruns  = 0;
    devasp_underruns = 0;
    devasp_dma_acnt  = 1;
    devasp_rsyncerr  = 0;
    devasp_wsyncerr  = 0;
  }

}


/******************************************************************************
  OsaLinuxAspSetRcvBufValue (Debug function)
    Fill output buffer with value
    return void
******************************************************************************/
void  
OsaLinuxAspFillOutBufValue(OsaUint8 value)
{
  int nIndex = 0;
  OsaUint8 *pBufAddr;

  for(nIndex = 0 ; nIndex < DEVASP_N_OUT_BUFFERS; nIndex++) {
    pBufAddr = (OsaUint8 *) devasp_out_buffers[nIndex].virt;
    if(pBufAddr != NULL) {
      memset(pBufAddr, value, DEVASP_BUF_SIZE);
    }
  }

}


/******************************************************************************
  OsaLinuxAspFillOutBufCount (Debug function)
    Fill output buffer with count sequence
    return void
******************************************************************************/
void  
OsaLinuxAspFillOutBufCount(OsaAspBufFillCntE type)
{
  OsaUint8  *pBufAddrByte;
  OsaUint16 *pBufAddrWord;
  OsaUint32 nIndex  = 0;
  OsaUint32 nBufCnt = 0;
  OsaUint32 nCnt    = 0;
  OsaUint32 nCntEnd = 0;

  OSA_DPRINTF(("OsaLinuxAspFillOutBufCount\n"));

  if(type == OSA_ASP_BYTE_CNT) {
    nCntEnd = 0xFF;
  }
  else
  if(type == OSA_ASP_WORD_CNT) {
    nCntEnd = 0xFFFF;
  }
  else {
    // exit nCntEnd = 0
    OSA_DPRINTF(("OsaLinuxAspFillOutBufCount: OsaAspBufFillCntE err\n"));
  }

  for(nIndex = 0 ; (nIndex < DEVASP_N_OUT_BUFFERS); nIndex++) {
    pBufAddrByte = (OsaUint8 *)  devasp_out_buffers[nIndex].virt;
    pBufAddrWord = (OsaUint16 *) devasp_out_buffers[nIndex].virt;
    if((pBufAddrByte != NULL) && (nCntEnd != 0)) {
      for(nBufCnt = 0; nBufCnt < DEVASP_BUF_SIZE; nBufCnt++) {
        if(type == OSA_ASP_BYTE_CNT) {
          *pBufAddrByte = (OsaUint8) nCnt;
        }
        else
        if(type == OSA_ASP_WORD_CNT) {
          *pBufAddrWord = (OsaUint16) nCnt;
        }
      }
      pBufAddrByte++;
      pBufAddrWord++;
      nCnt++;
      if(nCnt > nCntEnd) {
        nCnt = 0;
      }
    }
  }
}



//module_init(init_devasp);
//module_exit(fini_devasp);


#endif // DAVINCI
